/****************************************************************************
**
** Copyright (C) 2017 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the examples of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:BSD$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** BSD License Usage
** Alternatively, you may use this file under the terms of the BSD license
** as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of The Qt Company Ltd nor the names of its
**     contributors may be used to endorse or promote products derived
**     from this software without specific prior written permission.
**
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include "audiooutput.h"

#include <QAudioDeviceInfo>
#include <QAudioOutput>
#include <QDebug>
#include <QVBoxLayout>
#include <qmath.h>
#include <qendian.h>

//----------------------------------------------------
static int		LinToLogVol(qreal linVol)
{
	qreal	logVol(QAudio::convertVolume(
		linVol,
		QAudio::LinearVolumeScale, QAudio::LogarithmicVolumeScale) * 100);

	return qRound(logVol);
}

static qreal	LogToLinVol(int logVol)
{
	qreal	linVol = QAudio::convertVolume(
		logVol / qreal(100),
		QAudio::LogarithmicVolumeScale, QAudio::LinearVolumeScale);

	return linVol;
}

//----------------------------------------------------------------------------------
Generator::Generator(const QAudioFormat &format
    , qint64 durationUs
    , int sampleRate)
{
    if (format.isValid())
        generateData(format, durationUs, sampleRate);
}

void Generator::start()
{
    open(QIODevice::ReadOnly);
}

void Generator::stop()
{
    m_pos = 0;
    close();
}

void Generator::generateData(const QAudioFormat &format, qint64 durationUs, int sampleRate)
{
    const int channelBytes = format.sampleSize() / 8;
    const int sampleBytes = format.channelCount() * channelBytes;
    qint64 length = (format.sampleRate() * format.channelCount() * (format.sampleSize() / 8))
                        * durationUs / 1000000;
    Q_ASSERT(length % sampleBytes == 0);
    Q_UNUSED(sampleBytes) // suppress warning in release builds

    m_buffer.resize(length);
    unsigned char *ptr = reinterpret_cast<unsigned char *>(m_buffer.data());
    int sampleIndex = 0;

    while (length) {
        // Produces value (-1..1)
        const qreal x = qSin(2 * M_PI * sampleRate * qreal(sampleIndex++ % format.sampleRate()) / format.sampleRate());
        for (int i=0; i<format.channelCount(); ++i) {
            if (format.sampleSize() == 8) {
                if (format.sampleType() == QAudioFormat::UnSignedInt) {
                    const quint8 value = static_cast<quint8>((1.0 + x) / 2 * 255);
                    *reinterpret_cast<quint8 *>(ptr) = value;
                } else if (format.sampleType() == QAudioFormat::SignedInt) {
                    const qint8 value = static_cast<qint8>(x * 127);
                    *reinterpret_cast<qint8 *>(ptr) = value;
                }
            } else if (format.sampleSize() == 16) {
                if (format.sampleType() == QAudioFormat::UnSignedInt) {
                    quint16 value = static_cast<quint16>((1.0 + x) / 2 * 65535);
                    if (format.byteOrder() == QAudioFormat::LittleEndian)
                        qToLittleEndian<quint16>(value, ptr);
                    else
                        qToBigEndian<quint16>(value, ptr);
                } else if (format.sampleType() == QAudioFormat::SignedInt) {
                    qint16 value = static_cast<qint16>(x * 32767);
                    if (format.byteOrder() == QAudioFormat::LittleEndian)
                        qToLittleEndian<qint16>(value, ptr);
                    else
                        qToBigEndian<qint16>(value, ptr);
                }
            }

            ptr += channelBytes;
            length -= channelBytes;
        }
    }
}

qint64 Generator::readData(char *data, qint64 len)
{
    qint64 total = 0;
    if (!m_buffer.isEmpty()) {
        while (len - total > 0) {
            const qint64 chunk = qMin((m_buffer.size() - m_pos), len - total);
            memcpy(data + total, m_buffer.constData() + m_pos, chunk);
            m_pos = (m_pos + chunk) % m_buffer.size();
            total += chunk;
        }
    }
    return total;
}

qint64 Generator::writeData(const char *data, qint64 len)
{
    Q_UNUSED(data);
    Q_UNUSED(len);

    return 0;
}

qint64 Generator::bytesAvailable() const
{
    return m_buffer.size() + QIODevice::bytesAvailable();
}

//----------------------------------------------------------------------------------
AudioTest::AudioTest() :
	m_workerThread(this),
	m_audioWorkerP(NULL)
{
	m_audioWorkerP = new AudioWorker(this);

	connect(
		this,			&AudioTest::deviceChanged,
		m_audioWorkerP,	&AudioWorker::deviceChanged);

	//m_audioWorkerP->moveToThread(&m_workerThread);

	initializeWindow();

	connect(
		&m_workerThread,	&QThread::finished,
		m_audioWorkerP,		&QObject::deleteLater);

	m_workerThread.start();

	deviceMenuChanged(0);
}

AudioTest::~AudioTest()
{
	m_workerThread.quit();
	m_workerThread.wait();
}

void AudioTest::initializeWindow()
{
	QVBoxLayout		*layout(new QVBoxLayout);

	//	output device menu
	{
		m_deviceBox = new QComboBox(this);

		const QAudioDeviceInfo&		defaultDeviceInfo = QAudioDeviceInfo::defaultOutputDevice();

		m_deviceBox->addItem(defaultDeviceInfo.deviceName(), QVariant::fromValue(defaultDeviceInfo));

		for (auto &deviceInfo: QAudioDeviceInfo::availableDevices(QAudio::AudioOutput)) {

			if (deviceInfo != defaultDeviceInfo) {
				m_deviceBox->addItem(deviceInfo.deviceName(), QVariant::fromValue(deviceInfo));
			}
		}

		connect(
			m_deviceBox,	QOverload<int>::of(&QComboBox::activated),
			this,			&AudioTest::deviceMenuChanged);

		layout->addWidget(m_deviceBox);
	}

	//	push / pull mode button
	{
		m_modeButton = new QPushButton(this);

		connect(
			m_modeButton,	&QPushButton::clicked,
			m_audioWorkerP,	&AudioWorker::toggleMode);

		layout->addWidget(m_modeButton);
	}

	//	suspend / resume button
	{
		m_suspendResumeButton = new QPushButton(this);

		connect(
			m_suspendResumeButton,	&QPushButton::clicked,
			m_audioWorkerP,			&AudioWorker::toggleSuspendResume);

		layout->addWidget(m_suspendResumeButton);
	}

	//	volume slider
	{
		QHBoxLayout *volumeBox = new QHBoxLayout;
		m_volumeLabel = new QLabel;
		m_volumeLabel->setText(tr("Volume:"));
		m_volumeSlider = new QSlider(Qt::Horizontal);
		m_volumeSlider->setMinimum(0);
		m_volumeSlider->setMaximum(100);
		m_volumeSlider->setSingleStep(10);

		connect(
			m_volumeSlider,			&QSlider::valueChanged,
			m_audioWorkerP,			&AudioWorker::volumeChanged);

		volumeBox->addWidget(m_volumeLabel);
		volumeBox->addWidget(m_volumeSlider);
		layout->addLayout(volumeBox);
	}

	{
		QWidget			*window = new QWidget;

		window->setLayout(layout);

		setCentralWidget(window);
		window->show();
	}
}

void	AudioTest::updateControl(ControlType controlType, int valueI)
{
	switch (controlType) {

		case kControl_PUSH_PULL_BUTTON:
		case kControl_SUSPEND_RESUME_BUTTON: {
			QString			str;
			bool			push_pullB(controlType == kControl_PUSH_PULL_BUTTON);
			QPushButton		*buttonP = push_pullB ? m_modeButton : m_suspendResumeButton;

			//	kControl_PUSH_PULL_BUTTON,		then valueI == 1 means push, 0 means pull
			//	kControl_SUSPEND_RESUME_BUTTON, then valueI == 1 means suspend, 0 means resume

			if (push_pullB) {

				if (valueI) {
					//	current mode is PUSH
					str = tr("Enable Pull mode");
				} else {
					//	current mode is PULL
					str = tr("Enable Push mode");
				}
			} else {
				if (valueI) {
					//	current mode is SUSPEND
					str = tr("Resume");
				} else {
					//	current mode is RESUME
					str = tr("Suspend");
				}
			}

			buttonP->setText(str);

			//	this should not be necessary???
			buttonP->repaint();
		} break;

		case kControl_OUT_DEVICE_MENU: {
			//	 do we need to update the menu here?
		} break;

		case kControl_VOLUME_SLIDER: {
			//	valueI is log (UI view), not linear (audio hardware) volume
			m_volumeSlider->setValue(valueI);
		} break;
	}
}

void AudioTest::deviceMenuChanged(int valueI)
{
	const QAudioDeviceInfo&		devInfo(m_deviceBox->itemData(valueI).value<QAudioDeviceInfo>());

	emit deviceChanged(devInfo);
}

//----------------------------------------------------------------------------------
AudioWorker::AudioWorker(AudioTest *audioTestP) :
	m_pushTimer(new QTimer(this))
{
	m_curVolume = LogToLinVol(50);

	connect(
		this,		&AudioWorker::updateControl,
		audioTestP,	&AudioTest::updateControl,
		Qt::QueuedConnection);
}

AudioWorker::~AudioWorker()
{
	m_pushTimer->stop();
	delete m_pushTimer;
	m_pushTimer = nullptr;
}

void AudioWorker::initializeAudio(const QAudioDeviceInfo &deviceInfo)
{
	QAudioFormat format;

	format.setSampleRate(44100);
	format.setChannelCount(1);
	format.setSampleSize(16);
	format.setCodec("audio/pcm");
	format.setByteOrder(QAudioFormat::LittleEndian);
	format.setSampleType(QAudioFormat::SignedInt);

	if (!deviceInfo.isFormatSupported(format)) {
		qWarning() << "Default format not supported - trying to use nearest";
		format = deviceInfo.nearestFormat(format);
	}

	const int durationSeconds = 1;
	const int toneSampleRateHz = 600;

	m_generator.reset(new Generator(format, durationSeconds * 1000000, toneSampleRateHz));
	m_audioOutput.reset(new QAudioOutput(deviceInfo, format));
	m_generator->start();

	setPullMode(m_pullMode);
	setVolume(m_curVolume);
}

void AudioWorker::deviceChanged(QAudioDeviceInfo devInfo)
{
	if (m_generator) {
		m_generator->stop();
	}

	if (m_audioOutput) {
		m_audioOutput->stop();
		m_audioOutput->disconnect(this);
	}

	initializeAudio(devInfo);

	// ??? no need? or... ?
	// emit updateControl(kControl_OUT_DEVICE_MENU, index);
}

//----------------------------------------------------

void AudioWorker::setVolume(qreal linearVolume)
{
	m_curVolume = linearVolume;
	m_audioOutput->setVolume(linearVolume);

	int		logVolume(LinToLogVol(linearVolume));

	emit updateControl(kControl_VOLUME_SLIDER, logVolume);
}

void AudioWorker::volumeChanged(int logVolumeI)
{
	qreal		linearVolume(LogToLinVol(logVolumeI));

	setVolume(linearVolume);
}

void AudioWorker::setPullMode(bool pullModeB)
{
    m_pushTimer->stop();
    m_audioOutput->stop();

	setSuspendMode(false);

	m_pullMode = pullModeB;

	if (m_pullMode) {
		// switch to pull mode (QAudioOutput pulls from Generator as needed)
        m_audioOutput->start(m_generator.data());

    } else {
		// switch to push mode (periodically push to QAudioOutput using a timer)

		auto io = m_audioOutput->start();
        m_pushTimer->disconnect();

        connect(m_pushTimer, &QTimer::timeout, [this, io]() {
            if (m_audioOutput->state() == QAudio::StoppedState)
                return;

            QByteArray buffer(32768, 0);
            int chunks = m_audioOutput->bytesFree() / m_audioOutput->periodSize();
            while (chunks) {
               const qint64 len = m_generator->read(buffer.data(), m_audioOutput->periodSize());
               if (len)
                   io->write(buffer.data(), len);
               if (len != m_audioOutput->periodSize())
                   break;
               --chunks;
            }
        });

        m_pushTimer->start(20);
    }

	emit updateControl(kControl_PUSH_PULL_BUTTON, !m_pullMode);
}

void AudioWorker::toggleMode()
{
	setPullMode(!m_pullMode);
}

void AudioWorker::setSuspendMode(bool suspendB)
{
	if (suspendB) {
		m_audioOutput->suspend();
	} else {
		m_audioOutput->resume();
	}

	emit updateControl(kControl_SUSPEND_RESUME_BUTTON, suspendB);
}

void AudioWorker::toggleSuspendResume()
{
	bool		changeB = false;
	bool		suspendB = false;

    if (m_audioOutput->state() == QAudio::SuspendedState || m_audioOutput->state() == QAudio::StoppedState) {
		changeB = true;

    } else if (m_audioOutput->state() == QAudio::ActiveState) {
		changeB = true;
		suspendB = true;

    } else if (m_audioOutput->state() == QAudio::IdleState) {
        // no-op
    }

	if (changeB) {
		setSuspendMode(suspendB);
	}
}

